/*
 * EMEngine.cpp
 *
 *  Created on: Jan 19, 2010
 *      Author: duyvu
 */

#include "EMEngine.h"

#include "model/signed/MMExcessTrustModel.h"
#include "model/signed/NRMMExcessTrustModel.h"

#include "util/StringTokenizer.h"

#include <map>

#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cfloat>

namespace mixderg {

EMEngine::EMEngine() {
	numOfClasses = -1;
}

EMEngine::~EMEngine() {
	if (model != NULL)
		delete model;
}

void EMEngine::configure(const char* configurationFile, const char* delim) {

	std::cout << "The configuration file is " << configurationFile << std::endl;

	std::map<std::string, std::string> configurationMap;
	std::string line;
	std::ifstream fileStream(configurationFile);
	if (fileStream.is_open()) {
		while (!fileStream.eof()) {
			getline(fileStream, line);
			StringTokenizer strtok = StringTokenizer(line, delim);
			if (strtok.countTokens() >= 2) {
				std::string key = strtok.nextToken();
				std::string value = strtok.nextToken();
				configurationMap[key] = value;
			}
		}
	} else
		std::cout << "can not read the configuration file" << std::endl;

	for (std::map<std::string, std::string>::const_iterator it =
			configurationMap.begin(); it != configurationMap.end(); ++it) {
		std::cout << "(" << it->first << " , " << it->second << ")"
				<< std::endl;
	}

	modelType = atoi(configurationMap["Model_Type"].c_str());
	switch (modelType) {
	case NR_MM_EXCESS_TRUST_GRAPH:
		cout << "Creating NRMMExcessTrustModel" << endl;
		model = new NRMMExcessTrustModel();
		break;
	default:
		std::cout << "Need to specify the model type. I am quitting now!"
				<< std::endl;
		return;
	}

	if (numOfClasses == -1) {
		if (configurationMap.count("Num_Of_Clusters") == 1)
			numOfClasses = atoi(configurationMap["Num_Of_Clusters"].c_str());
		else
			numOfClasses = 2;
	}
	model->setNumOfClasses(numOfClasses);

	if (configurationMap.count("Data_File") == 1
			&& configurationMap.count("Data_Format") == 1) {
		const char* dataFile = configurationMap["Data_File"].c_str();
		int dataFormat = atoi(configurationMap["Data_Format"].c_str());
		std::string delim;
		switch (atoi(configurationMap["Delim"].c_str())) {
		case BLANK_SPACE:
			delim = " ";
			break;
		case TAB:
			delim = "\t";
			break;
		default:
			delim = " ";
			break;
		}
		model->setData(dataFile, dataFormat, delim.c_str());

		if (configurationMap.count("SEED_1") == 1)
			seed1 = atoi(configurationMap["SEED_1"].c_str());
		else
			seed1 = 45;

		if (configurationMap.count("SEED_2") == 1)
			seed2 = atoi(configurationMap["SEED_2"].c_str());
		else
			seed2 = 345;

		if (configurationMap.count("Num_Of_EM_Runs") == 1)
			numOfEMRuns = atoi(configurationMap["Num_Of_EM_Runs"].c_str());
		else
			numOfEMRuns = 1;

		if (configurationMap.count("Max_Num_Of_EM_Iterations") == 1)
			maxNumOfEMIterations = atoi(
					configurationMap["Max_Num_Of_EM_Iterations"].c_str());
		else
			maxNumOfEMIterations = 30;

		if (configurationMap.count("Max_Num_Of_E_Steps") == 1)
			maxNumOfESteps = atoi(
					configurationMap["Max_Num_Of_E_Steps"].c_str());
		else
			maxNumOfESteps = 30;

		if (configurationMap.count("Min_Tau") == 1)
			minTau = atof(configurationMap["Min_Tau"].c_str());
		else
			minTau = 1e-10;
		model->setMinTau(minTau);

		if (configurationMap.count("Min_Alpha") == 1)
			minAlpha = atof(configurationMap["Min_Alpha"].c_str());
		else
			minAlpha = 1e-10;
		model->setMinAlpha(minAlpha);

		if (configurationMap.count("Tau_Precision") == 1)
			tauPrecision = atof(configurationMap["Tau_Precision"].c_str());
		else
			tauPrecision = 1e-10;

		if (configurationMap.count("ParamPrecision") == 1)
			paramPrecision = atof(configurationMap["ParamPrecision"].c_str());
		else
			paramPrecision = 1e-10;
	} else {
		std::cout << "No data available. I am quitting now!" << std::endl;
		return;
	}
}

void EMEngine::setData(const char* dataFile, int fileFormat,
		const char* delim) {
	model->setData(dataFile, fileFormat, delim);
}

void EMEngine::runEM() {
	std::cout << "EMEngine::runEM" << std::endl;

	cout.setf(ios::fixed, ios::floatfield);
	cout.precision(20);

	// Assuming that model has been created
	// data and all parameters have been configured
	model->resetParameters();

	std::cout << "Starting a EM run" << std::endl;
	for (int emIteration = 0; emIteration < maxNumOfEMIterations;
			emIteration++) {

		std::cout << "EM Iteration " << emIteration << std::endl;

		for (int eStep = 0; eStep < maxNumOfESteps; eStep++) {
			std::cout << "E Step " << eStep << std::endl;
			model->runFixedPointEstimationEStep(emIteration, eStep);
			if (!model->isTauSignificantlyChanged(tauPrecision))
				break;
		}

		model->runModelEstimationMStep(emIteration);
		if (!model->areAlphaPiSignificantlyChanged(paramPrecision)) {
			cout << "alpha and pi are not significantly changed!" << endl;
			break;
		}
	}
}

void EMEngine::runDetailedEM() {
	cout << "EMEngine::runDetailedEM" << endl;

	cout.setf(ios::fixed, ios::floatfield);
	cout.precision(20);

	// Assuming that model has been created
	// Data and all parameters have been configured
	model->resetParameters();

	MMExcessTrustModel* myModel = (MMExcessTrustModel*) model;

	cout << "Starting a EM run" << endl;
	double lastLLHLowerBound = 0;
	for (int emIteration = 0; emIteration < maxNumOfEMIterations;
			emIteration++) {

		// E Step
		for (int eStep = 0; eStep < maxNumOfESteps; eStep++)
			model->runFixedPointEstimationEStep(emIteration, eStep);

		// M Step
		model->runModelEstimationMStep(emIteration);

		// Covergence Details
		double newLLHLowerBound = model->getLLHLowerBound();
		double largestTauChange = model->getLargestTauChange();
		cout << emIteration << "\t";
		cout << newLLHLowerBound << "\t";
		cout << largestTauChange << "\t";
		for (unsigned int k = 0; k < model->numOfClasses; k++)
			cout << myModel->excessTrustTheta[k] << "\t";
		cout << myModel->negDenTheta << "\t";
		cout << myModel->negRecTheta << "\t";
		cout << myModel->posRecTheta << endl;

		// Convergence Criterion Check
		if ((emIteration > 0)
				&& (fabs(
						(newLLHLowerBound - lastLLHLowerBound)
								/ newLLHLowerBound) < REL_LLH))
			break;
		else
			lastLLHLowerBound = newLLHLowerBound;
	}
}

void EMEngine::runInitializedEM(const char* tauFile, const char* betaFile) {
	std::cout << "EMEngine::runInitializedEM" << std::endl;

	cout.setf(ios::fixed, ios::floatfield);
	cout.precision(20);

	// Finalize all configuration setttings
	model->setModel();

	// Initialize parameters
	string line;

	MMExcessTrustModel* myModel = (MMExcessTrustModel*) model;

	ifstream tauStream(tauFile);
	if (tauStream.is_open()) {
		if (!tauStream.eof()) {
			for (unsigned int i = 0; i < model->numOfVertices; i++) {
				getline(tauStream, line);
				StringTokenizer strtok = StringTokenizer(line, "\t");
				for (unsigned int k = 0; k < model->numOfClasses; k++)
					myModel->tau[i][k] = strtok.nextDoubleToken();
			}
			tauStream.close();
		}
	} else
		cout << "Unable to open tau file " << tauFile << endl;

	ifstream betaStream(betaFile);
	if (betaStream.is_open()) {
		if (!betaStream.eof()) {
			// alpha
			getline(betaStream, line);
			getline(betaStream, line);
			StringTokenizer strtok = StringTokenizer(line, "\t");
			for (unsigned int k = 0; k < model->numOfClasses; k++)
				myModel->alpha[k] = strtok.nextDoubleToken();

			// excessTrustTheta:
			getline(betaStream, line);
			getline(betaStream, line);
			strtok = StringTokenizer(line, "\t");
			for (unsigned int k = 0; k < model->numOfClasses; k++)
				myModel->excessTrustTheta[k] = strtok.nextDoubleToken();

			// negDenTheta:
			getline(betaStream, line);
			getline(betaStream, line);
			strtok = StringTokenizer(line, "\t");
			myModel->negDenTheta = strtok.nextDoubleToken();

			// negRecTheta:
			getline(betaStream, line);
			getline(betaStream, line);
			strtok = StringTokenizer(line, "\t");
			myModel->negRecTheta = strtok.nextDoubleToken();

			// posRecTheta:
			getline(betaStream, line);
			getline(betaStream, line);
			strtok = StringTokenizer(line, "\t");
			myModel->posRecTheta = strtok.nextDoubleToken();

			betaStream.close();
		}
	} else
		cout << "Unable to open beta file " << betaFile << endl;

	std::cout << "Starting a EM run" << std::endl;
	for (int emIteration = 0; emIteration < maxNumOfEMIterations;
			emIteration++) {

		std::cout << "EM Iteration " << emIteration << std::endl;

		for (int eStep = 0; eStep < maxNumOfESteps; eStep++) {
			std::cout << "E Step " << eStep << std::endl;
			model->runFixedPointEstimationEStep(emIteration, eStep);
			if (!model->isTauSignificantlyChanged(tauPrecision)) {
				cout << "tau is not significantly changed!" << endl;
				break;
			}
		}

		model->runModelEstimationMStep(emIteration);
		if (!model->areAlphaPiSignificantlyChanged(paramPrecision)) {
			cout << "alpha and pi are not significantly changed!" << endl;
			break;
		}

		cout << "Current Lower Bound = " << model->getLLHLowerBound() << endl;
		printDoubleVector("excessTrustTheta: ", myModel->excessTrustTheta);
		cout << "negDenTheta: " << myModel->negDenTheta << endl;
		cout << "negRecTheta: " << myModel->negRecTheta << endl;
		cout << "posRecTheta: " << myModel->posRecTheta << endl;
	}
}

void EMEngine::runOneModel(string modelLLHFilename, string bestModelFilename,
		int outputPrecision) {

	// Finalize all configuration setttings
	model->setModel();

	double bestLowerBound = -DBL_MAX;
	for (int run = 0; run < numOfEMRuns; run++) {

		// Run EM algorithm and compute the lower bound achieved
		runEM();
		double currentLowerBound = model->getLLHLowerBound();

		// Log to console the current run
		cout << "Run " << run << " achieved a Likelihood Lower Bound = "
				<< currentLowerBound << endl;

		// Output the lower bound of the current run for model selection
		ofstream modelLLHFile;
		modelLLHFile.open(modelLLHFilename.c_str(), ios::app);
		modelLLHFile.setf(ios::fixed, ios::floatfield);
		modelLLHFile.precision(outputPrecision);
		modelLLHFile << run << "\t" << currentLowerBound << endl;
		modelLLHFile.close();

		// Check if the lower bound is improved
		if (bestLowerBound < currentLowerBound) {
			bestLowerBound = currentLowerBound;
			// Prepare the output file
			ofstream bestModelFile;
			bestModelFile.open(bestModelFilename.c_str(), ios::trunc);
			bestModelFile.setf(ios::fixed, ios::floatfield);
			bestModelFile.precision(outputPrecision);
			bestModelFile << bestLowerBound << endl;
			model->writeModel2File(bestModelFile);
			bestModelFile.close();
		}
	}
}

void EMEngine::runOneModel(string modelLLHFilename, string bestModelFilename,
		string bestLLHFilename, int outputPrecision, int fromRun) {

	// Finalize all configuration setttings
	model->setModel();

	// This object will contain parameters of the best model
	DegreeModel* bestModel = NULL;
	switch (modelType) {
	case NR_MM_EXCESS_TRUST_GRAPH:
		cout << "Creating a copy model in runOneModel" << endl;
		bestModel = new NRMMExcessTrustModel();
		break;
	default:
		cout << "The MODEL TYPE need to be specified!!!" << endl;
		return;
	}

	double bestLLH = -DBL_MAX;
	for (int run = 0; run < numOfEMRuns; run++) {

		// Run EM algorithm and compute the lower bound achieved
		runEM();
		double currentLLH = model->getLLHLowerBound();

		// Log to console the current run
		cout << "Run " << run << " achieved a LLH = " << currentLLH << endl;

		// Output the lower bound of the current run for model selection
		ofstream modelLLHFile;
		modelLLHFile.open(modelLLHFilename.c_str(), ios::out | ios::app);
		modelLLHFile.setf(ios::fixed, ios::floatfield);
		modelLLHFile.precision(outputPrecision);
		modelLLHFile << (fromRun + run) << "\t" << currentLLH << endl;
		modelLLHFile.close();

		// Check if the lower bound is improved
		if (bestLLH < currentLLH) {
			bestLLH = currentLLH;
			bestModel->copyParameters(model); // copy the current model to bestModel
		}
	}

	ifstream readBestLLHFile;
	readBestLLHFile.open(bestLLHFilename.c_str());
	if (readBestLLHFile.is_open()) {
		double currentBestLLH;
		readBestLLHFile >> currentBestLLH;
		readBestLLHFile.close();
		if (currentBestLLH < bestLLH) {
			// Overwrite the old current best ICL
			ofstream writeBestLLHFile;
			writeBestLLHFile.open(bestLLHFilename.c_str());
			if (writeBestLLHFile.is_open()) {
				writeBestLLHFile.precision(outputPrecision);
				writeBestLLHFile << bestLLH;
				writeBestLLHFile.close();
			} else
				cout << "Can not open the file " << bestLLHFilename << endl;
			// Update the best model
			ofstream bestModelFile;
			bestModelFile.open(bestModelFilename.c_str());
			bestModelFile.setf(ios::fixed, ios::floatfield);
			bestModelFile.precision(outputPrecision);
			bestModel->writeModel2File(bestModelFile);
			bestModelFile.close();
		}
	} else {
		ofstream writeBestLLHFile;
		writeBestLLHFile.open(bestLLHFilename.c_str());
		if (writeBestLLHFile.is_open()) {
			// Write the current best ICL
			writeBestLLHFile.precision(outputPrecision);
			writeBestLLHFile << bestLLH;
			writeBestLLHFile.close();
			// Update the best model
			ofstream bestModelFile;
			bestModelFile.open(bestModelFilename.c_str());
			bestModelFile.setf(ios::fixed, ios::floatfield);
			bestModelFile.precision(outputPrecision);
			bestModel->writeModel2File(bestModelFile);
			bestModelFile.close();
		} else
			cout << "Can not open the file " << bestLLHFilename << endl;
	}

	if (bestModel != NULL)
		delete bestModel;
}

void EMEngine::runDetailedOneModel(string modelLLHFilename,
		string bestModelFilename, string bestLLHFilename, int outputPrecision,
		int fromRun) {

	// Finalize all configuration setttings
	model->setModel();

	// This object will contain parameters of the best model
	DegreeModel* bestModel = NULL;
	switch (modelType) {
	case NR_MM_EXCESS_TRUST_GRAPH:
		cout << "Creating copying model" << endl;
		bestModel = new NRMMExcessTrustModel();
		break;
	default:
		cout << "The MODEL TYPE need to be specified!!!" << endl;
		return;
	}

	cout << "Done copying model" << endl;
	cout << "num EM Runs = " << numOfEMRuns << endl;

	double bestLLH = -DBL_MAX;
	for (int run = 0; run < numOfEMRuns; run++) {

		// Run EM algorithm and compute the lower bound achieved
		runDetailedEM();
		double currentLLH = model->getLLHLowerBound();

		// Log to console the current run
		cout << "Run " << run << " achieved a LLH = " << currentLLH << endl;

		// Output the lower bound of the current run for model selection
		ofstream modelLLHFile;
		modelLLHFile.open(modelLLHFilename.c_str(), ios::out | ios::app);
		modelLLHFile.setf(ios::fixed, ios::floatfield);
		modelLLHFile.precision(outputPrecision);
		modelLLHFile << (fromRun + run) << "\t" << currentLLH << endl;
		modelLLHFile.close();

		// Check if the lower bound is improved
		if (bestLLH < currentLLH) {
			bestLLH = currentLLH;
			bestModel->copyParameters(model); // copy the current model to bestModel
		}
	}

	ifstream readBestLLHFile;
	readBestLLHFile.open(bestLLHFilename.c_str());
	if (readBestLLHFile.is_open()) {
		double currentBestLLH;
		readBestLLHFile >> currentBestLLH;
		readBestLLHFile.close();
		if (currentBestLLH < bestLLH) {
			// Overwrite the old current best ICL
			ofstream writeBestLLHFile;
			writeBestLLHFile.open(bestLLHFilename.c_str());
			if (writeBestLLHFile.is_open()) {
				writeBestLLHFile.precision(outputPrecision);
				writeBestLLHFile << bestLLH;
				writeBestLLHFile.close();
			} else
				cout << "Can not open the file " << bestLLHFilename << endl;
			// Update the best model
			ofstream bestModelFile;
			bestModelFile.open(bestModelFilename.c_str());
			bestModelFile.setf(ios::fixed, ios::floatfield);
			bestModelFile.precision(outputPrecision);
			bestModel->writeModel2File(bestModelFile);
			bestModelFile.close();
		}
	} else {
		ofstream writeBestLLHFile;
		writeBestLLHFile.open(bestLLHFilename.c_str());
		if (writeBestLLHFile.is_open()) {
			// Write the current best ICL
			writeBestLLHFile.precision(outputPrecision);
			writeBestLLHFile << bestLLH;
			writeBestLLHFile.close();
			// Update the best model
			ofstream bestModelFile;
			bestModelFile.open(bestModelFilename.c_str());
			bestModelFile.setf(ios::fixed, ios::floatfield);
			bestModelFile.precision(outputPrecision);
			bestModel->writeModel2File(bestModelFile);
			bestModelFile.close();
		} else
			cout << "Can not open the file " << bestLLHFilename << endl;
	}

	if (bestModel != NULL)
		delete bestModel;
}

}
