/*
 * DegreeModel.cpp
 *
 *  Created on: Oct 9, 2011
 *      Author: duyvu
 */

#include "DegreeModel.h"

#include "cmath"

namespace mixderg {

DegreeModel::DegreeModel() {
	numOfVertices = 0;
	numOfEdges = 0;
	numOfClasses = 0;
}

DegreeModel::~DegreeModel() {
}

void DegreeModel::sumDoubleMatrixOverRows(const DoubleMatrix & matrix,
		DoubleVector & vector) {
	if (matrix.size() > 0) {
		for (unsigned int k = 0; k < vector.size(); k++)
			vector[k] = 0;
		for (unsigned int i = 0; i < matrix.size(); i++)
			for (unsigned int k = 0; k < vector.size(); k++)
				vector[k] += matrix[i][k];
	}
}

void DegreeModel::normalizeTau(DoubleMatrix& tau, double minValue) {
	// normalize
	for (unsigned int i = 0; i < numOfVertices; i++) {

		double denominator = 0;

		for (unsigned int k = 0; k < numOfClasses; k++)
			denominator += tau[i][k];

		bool again = false;
		for (unsigned int k = 0; k < numOfClasses; k++) {
			tau[i][k] /= denominator;
			if (tau[i][k] < minValue) {
				tau[i][k] = minValue;
				again = true;
			}
		}

		if (again) {
			denominator = 0;
			for (unsigned int k = 0; k < numOfClasses; k++)
				denominator += tau[i][k];
			for (unsigned int k = 0; k < numOfClasses; k++)
				tau[i][k] /= denominator;
		}
	}
}

void DegreeModel::normalizeVector(DoubleVector& vector, double minValue) {
	for (unsigned int k = 0; k < vector.size(); k++)
		if (vector[k] < minValue)
			vector[k] = minValue;
	double sumAlpha = 0;
	for (unsigned int k = 0; k < vector.size(); k++)
		sumAlpha += vector[k];
	for (unsigned int k = 0; k < vector.size(); k++)
		vector[k] /= sumAlpha;
}

/* Solve QP with one equality constraint and bounded constraints */
void DegreeModel::solveQP(const DoubleMatrix& m, const DoubleMatrix& s,
		DoubleMatrix& tau, double precision) {

	if (m.size() <= 0)
		return;

	unsigned int n = m[0].size();

	bool J_k[n];
	bool J_lambda_a[n];

	bool J_lambda_k[n];
	bool J_lambda_k_a[n];
	bool J_lambda_k_b[n];

	double lambda_k;

	for (unsigned int rowIndex = 0; rowIndex < m.size(); rowIndex++) {
		// Step 1:
		for (unsigned int j = 0; j < n; j++) {
			J_k[j] = true;
			J_lambda_a[j] = false;
		}

		bool done = false;
		// we never loop over n times.
		unsigned int count = 0;
		while (!done) {
			count++;

			// Step 2:
			double value1 = 0, value2 = 0;
			for (unsigned int j = 0; j < n; j++)
				if (J_k[j]) {
					value1 += 1 / m[rowIndex][j];
					value2 += s[rowIndex][j] / m[rowIndex][j];
				}
			lambda_k = (value2 - 2) / value1;

			// Step3:
			for (unsigned int j = 0; j < n; j++) {
				if (J_k[j]) {
					if (lambda_k >= s[rowIndex][j]) {
						J_lambda_k[j] = false;
						J_lambda_k_a[j] = true;
						J_lambda_k_b[j] = false;
					} else if (lambda_k
							> (-2 * m[rowIndex][j] + s[rowIndex][j])) {
						J_lambda_k[j] = true;
						J_lambda_k_a[j] = false;
						J_lambda_k_b[j] = false;
					} else {
						J_lambda_k[j] = false;
						J_lambda_k_a[j] = false;
						J_lambda_k_b[j] = true;
					}
				} else {
					J_lambda_k[j] = false;
					J_lambda_k_a[j] = false;
					J_lambda_k_b[j] = false;
				}
			}

			// Step 4:
			double delta = 0;
			double value3 = 0, value4 = 0;
			bool J_lambda_k_empty = true;
			for (unsigned int j = 0; j < n; j++) {
				delta += J_lambda_k_b[j];
				if (J_lambda_k[j]) {
					value3 += s[rowIndex][j] / m[rowIndex][j];
					value4 += 1 / m[rowIndex][j];
					J_lambda_k_empty = false;
				}
			}
			delta += (value3 / 2 - lambda_k * value4 / 2 - 1);

			if (fabs(delta) < precision || J_lambda_k_empty || count >= n) {
				for (unsigned int j = 0; j < n; j++) {
					if (J_lambda_a[j] || J_lambda_k_a[j])
						tau[rowIndex][j] = 0;
					else if (J_lambda_k_b[j])
						tau[rowIndex][j] = 1;
					else
						tau[rowIndex][j] = (s[rowIndex][j] - lambda_k)
								/ (2 * m[rowIndex][j]);
				}
				done = true;
			} else if (delta > 0) {
				for (unsigned int j = 0; j < n; j++) {
					if (J_lambda_k_a[j]) {
						J_lambda_a[j] = true;
						J_k[j] = false;
					}
				}
			} else {
				for (unsigned int j = 0; j < n; j++) {
					if (J_lambda_k_b[j])
						tau[rowIndex][j] = 1;
					else
						tau[rowIndex][j] = 0;
				}
				done = true;
			}
		}
	}
}

}
