/*
 * MMExcessTrustModel.cpp
 *
 *  Created on: Dec 9, 2011
 *      Author: duyvu
 */

#include "MMExcessTrustModel.h"

#include "util/StringTokenizer.h"
#include "model/DataTypes.h"

#include "cmath"

#define MATHLIB_STANDALONE
#include <Rmath.h>

using namespace std;

namespace mixderg {

MMExcessTrustModel::MMExcessTrustModel() {
}

MMExcessTrustModel::~MMExcessTrustModel() {
}

void MMExcessTrustModel::computeNetworkStatistics(int i, int j, int value,
		VectorOfNodeSet & negFollowers, VectorOfNodeSet & posFollowers) {
	excessTrustDegrees[j] += value;
	NodeSet::iterator it;
	switch (value) {
	case -1:
		negDensity++;
		it = negFollowers[i].find(j);
		if (it != negFollowers[i].end()) {
			negFollowers[i].erase(it);
			negReciprocity++;
		} else
			negFollowers[j].insert(i);
		break;
	case 1:
		it = posFollowers[i].find(j);
		if (it != posFollowers[i].end()) {
			posFollowers[i].erase(it);
			posReciprocity++;
		} else
			posFollowers[j].insert(i);
		break;
	}
}

void MMExcessTrustModel::setData(const char* dataFile, int fileFormat,
		const char* delim) {

	cout << "The network data file is " << dataFile << endl;

	string line;
	ifstream fileStream(dataFile);
	if (fileStream.is_open()) {
		// The first line is the number of vertices
		if (!fileStream.eof()) {
			// Get the number of vertices
			getline(fileStream, line);
			StringTokenizer strtok = StringTokenizer(line, delim);
			numOfVertices = strtok.nextIntToken();
			cout << "The number of vertices is " << numOfVertices << endl;

			// Create and reset network statistics
			excessTrustDegrees = DoubleVector(numOfVertices, 0);
			negDensity = 0;
			negReciprocity = 0;
			posReciprocity = 0;

			// The following data structure will allow us to compute - and + reciprocity online
			VectorOfNodeSet negFollowers = VectorOfNodeSet(numOfVertices,
					NodeSet());
			VectorOfNodeSet posFollowers = VectorOfNodeSet(numOfVertices,
					NodeSet());

			// We will count the number of valid edges during reading
			numOfEdges = 0;

			// For sparse format matrix
			if (fileFormat == SPARSE_FORMAT) {
				cout << "SPARSE_FORMAT" << endl;
				while (!fileStream.eof()) {
					getline(fileStream, line);
					//cout << line << endl;
					StringTokenizer strtok = StringTokenizer(line, delim);
					//cout << strtok.countTokens() << endl;
					if (strtok.countTokens() == 3) {
						int i = strtok.nextIntToken();
						int j = strtok.nextIntToken();
						if (i == j) // self-edge will be removed
							continue;
						int value = strtok.nextIntToken();
						//cout << "(" << i << ", " << j << ") = " << value << endl;
						numOfEdges++;
						computeNetworkStatistics(i, j, value, negFollowers,
								posFollowers);
					}
				}
			}

			// For normal matrix
			if (fileFormat == MATRIX_FORMAT) {
				cout << "MATRIX_FORMAT" << endl;
				int i = 0;
				while (!fileStream.eof()) {
					getline(fileStream, line);
					StringTokenizer strtok = StringTokenizer(line, delim);
					//cout << strtok.countTokens() << endl;
					if (strtok.countTokens() == numOfVertices) {
						for (unsigned int j = 0; j < numOfVertices; j++) {
							int value = strtok.nextIntToken();
							if (value != 0) {
								//cout << "(" << i << ", " << j << ") = " << value << endl;
								numOfEdges++;
								computeNetworkStatistics(i, j, value,
										negFollowers, posFollowers);
							}
						}
					}
					i++;
				}
			}

			cout << "The number of edges is " << numOfEdges << endl;

//			cout << "excessTrustDegrees = ";
//			for (unsigned int i = 0; i < numOfVertices; i++)
//				cout << excessTrustDegrees[i] << ", ";
//			cout << endl;
//			cout << "negDensity = " << negDensity << endl;
//			cout << "negReciprocity = " << negReciprocity << endl;
//			cout << "posReciprocity = " << posReciprocity << endl;
		}

		fileStream.close();
	} else
		cout << "Unable to open file " << dataFile << endl;
}

void MMExcessTrustModel::setModel() {

	cout << "MMExcessTrustModel::setModel()" << endl;
	cout << "Number Of Vertices = " << numOfVertices << endl;
	cout << "Number Of Classes = " << numOfClasses << endl;

	cout << "Creating parameter matrices" << endl;
	tau = DoubleMatrix(numOfVertices, DoubleVector(numOfClasses));
	prevTau = DoubleMatrix(numOfVertices, DoubleVector(numOfClasses));

	alpha = DoubleVector(numOfClasses);
	prevAlpha = DoubleVector(numOfClasses);

	excessTrustTheta = DoubleVector(numOfClasses);
	prev_excessTrustTheta = DoubleVector(numOfClasses);

	// Gamma terms
	kappa = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));
	gamma = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));
	gamma1 = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));
	gamma2 = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));
	gamma3 = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));
	gamma4Minus = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));
	gamma4Plus = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));
}

void MMExcessTrustModel::setModel(int _numOfClasses, double _minTau,
		double _minAlpha, double _minTheta) {
	cout << "MMExcessTrustModel::setModel(...)" << endl;
	// set up basic information
	cout << "Setting up basic information" << endl;
	setNumOfClasses(_numOfClasses);
	minTau = _minTau;
	minAlpha = _minAlpha;
	minTheta = _minTheta;
	setModel();
}

void MMExcessTrustModel::resetParameters() {
	// set up parameters
	cout << "Resetting tau randomly" << endl;

	cout << "Vertices = " << numOfVertices << endl;
	cout << "Classes = " << numOfClasses << endl;

	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int k = 0; k < numOfClasses; k++)
			tau[i][k] = runif(0, 1);
	normalizeTau(tau, minTau);

	cout << "Resetting model parameters" << endl;
	for (unsigned int k = 0; k < numOfClasses; k++) {
		alpha[k] = 0;
		excessTrustTheta[k] = 0;
	}
	negDenTheta = 0;
	negRecTheta = 0;
	posRecTheta = 0;
	updateKappa();

	runModelEstimationMStep(0);

	cout << "Initial Lower Bound is " << getLLHLowerBound() << endl;
}

void MMExcessTrustModel::resetParameters(const UnsignedIntVector& Z) {
	// set up parameters
	cout << "Resetting tau using Z" << endl;

	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int k = 0; k < numOfClasses; k++)
			tau[i][k] = 0;

	cout << "Using HARD assignment" << endl;
	for (unsigned int i = 0; i < numOfVertices; i++)
		tau[i][Z[i]] = 1;
	normalizeTau(tau, minTau);

	cout << "Resetting model parameters" << endl;
	for (unsigned int k = 0; k < numOfClasses; k++) {
		alpha[k] = 0;
		excessTrustTheta[k] = 0;
	}
	negDenTheta = 0;
	negRecTheta = 0;
	posRecTheta = 0;
	updateKappa();

	runModelEstimationMStep(0);
}

void MMExcessTrustModel::resetParametersPartially(const UnsignedIntVector& Z,
		double hardPercentage) {
	// set up parameters
	cout << "Resetting tau using Z PARTIALLY with HARD_PERCENTAGE = "
			<< hardPercentage << endl;

	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int k = 0; k < numOfClasses; k++)
			tau[i][k] = 0;

	cout << "Using SOFT assignment" << endl;
	for (unsigned int i = 0; i < numOfVertices; i++)
		if (runif(0, 1) < hardPercentage)
			tau[i][Z[i]] = 1;
		else
			for (unsigned int k = 0; k < numOfClasses; k++)
				if (k != Z[i])
					tau[i][k] = runif(0, 1);
	normalizeTau(tau, minTau);

	cout << "Resetting model parameters" << endl;
	for (unsigned int k = 0; k < numOfClasses; k++) {
		alpha[k] = 0;
		excessTrustTheta[k] = 0;
	}
	negDenTheta = 0;
	negRecTheta = 0;
	posRecTheta = 0;
	updateKappa();

	runModelEstimationMStep(0);
}

void MMExcessTrustModel::resetParameters(const DoubleMatrix& initTau) {
	// set up parameters
	cout << "Resetting tau using initTau" << endl;

	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int k = 0; k < numOfClasses; k++)
			tau[i][k] = initTau[i][k];
	normalizeTau(tau, minTau);

	cout << "Resetting model parameters" << endl;
	for (unsigned int k = 0; k < numOfClasses; k++) {
		alpha[k] = 0;
		excessTrustTheta[k] = 0;
	}
	negDenTheta = 0;
	negRecTheta = 0;
	posRecTheta = 0;
	updateKappa();

	runModelEstimationMStep(0);
}

void MMExcessTrustModel::runFixedPointEstimationEStep(int emIter, int eIter) {

	cout << "MMExcessTrustModel::runFixedPointEstimationEStep()" << endl;

	static DoubleMatrix A = DoubleMatrix(numOfVertices,
			DoubleVector(numOfClasses));
	static DoubleMatrix s = DoubleMatrix(numOfVertices,
			DoubleVector(numOfClasses));

	// calculate the quadratic coefficients, i.e. A(i,k)
	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int k = 0; k < numOfClasses; k++) {
			double term = 0;
			for (unsigned int l = 0; l < numOfClasses; l++)
				term += kappa[k][l] * (numOfVertices * alpha[l] - tau[i][l]);
			A[i][k] = (1.0 + term / 2) / tau[i][k];
		}

	// calculate the linear coefficients, i.e. s(i,k)
	DoubleVector logAlpha = DoubleVector(numOfClasses);
	for (unsigned int k = 0; k < numOfClasses; k++)
		logAlpha[k] = log(alpha[k]);
	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int k = 0; k < numOfClasses; k++)
			s[i][k] = 1 + logAlpha[k] - log(tau[i][k])
					+ excessTrustDegrees[i] * excessTrustTheta[k];

	DoubleMatrix new_tau = DoubleMatrix(numOfVertices,
			DoubleVector(numOfClasses));
	solveQP(A, s, new_tau, minTau);

	normalizeTau(new_tau, minTau);

	prevTau.swap(tau);
	tau.swap(new_tau);
}

bool MMExcessTrustModel::isTauSignificantlyChanged(double _tauPrecision) {
	return false; // since we only update tau one time in the variational E step
}

double MMExcessTrustModel::getLargestTauChange() {
	double largestChange = 0;
	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int k = 0; k < numOfClasses; k++)
			if (fabs(tau[i][k] - prevTau[i][k]) > largestChange)
				largestChange = fabs(tau[i][k] - prevTau[i][k]);
	return largestChange;
}

double MMExcessTrustModel::computeCarryOverTerm(const DoubleMatrix & delta) {
	double norm = 0;
	for (unsigned int l = 0; l < numOfClasses; l++) {
		double carryTerm = 0;
		for (unsigned int j = 1; j < numOfVertices; j++) {
			for (unsigned k = 0; k < numOfClasses; k++)
				carryTerm += tau[j - 1][k] * delta[k][l];
			norm += tau[j][l] * carryTerm;
		}
	}
	return norm;
}

double MMExcessTrustModel::computeCarryOverTermByClass(unsigned int classIndex,
		const DoubleMatrix & delta) {
	double norm = 0;
	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int j = 0; j < i; j++)
			for (unsigned int h = 0; h < numOfClasses; h++)
				norm += tau[i][classIndex] * tau[j][h] * delta[classIndex][h]
						+ tau[i][h] * tau[j][classIndex] * delta[h][classIndex];
	return norm;
}

double MMExcessTrustModel::computeWeightedDegrees(unsigned int classIndex,
		const DoubleVector & degrees) {
	double term = 0;
	for (unsigned int i = 0; i < numOfVertices; i++)
		term += tau[i][classIndex] * degrees[i];
	return term;
}

bool MMExcessTrustModel::areAlphaPiSignificantlyChanged(
		double _paramPrecision) {

	cout << "MMExcessTrustModel::areAlphaPiSignificantlyChanged" << endl;

	for (unsigned int k = 0; k < numOfClasses; k++)
		if (fabs(alpha[k] - prevAlpha[k]) > _paramPrecision)
			return true;

	for (unsigned int k = 0; k < numOfClasses; k++)
		if (fabs(excessTrustTheta[k] - prev_excessTrustTheta[k])
				> _paramPrecision)
			return true;

	if (fabs(negDenTheta - prev_negDenTheta) > _paramPrecision)
		return true;

	if (fabs(negRecTheta - prev_negRecTheta) > _paramPrecision)
		return true;

	if (fabs(posRecTheta - prev_posRecTheta) > _paramPrecision)
		return true;

	return false;
}

double MMExcessTrustModel::getLLHLowerBound() {

	updateKappa();

	double CLL = 0;
	// tau * [ log(alpha) - log(tau) + pos in degree statistics ]
	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int k = 0; k < numOfClasses; k++)
			CLL += tau[i][k]
					* (log(alpha[k] / tau[i][k])
							+ excessTrustDegrees[i] * excessTrustTheta[k]);
	// density statistics
	CLL += negDenTheta * negDensity;
	// reciprocity statistics
	CLL += negRecTheta * negReciprocity;
	CLL += posRecTheta * posReciprocity;
	// norm term
	CLL -= computeCarryOverTerm(kappa);
	return CLL;
}

double MMExcessTrustModel::getICL() {
	double CLL = getLLHLowerBound();
	double ICL = CLL - 0.5 * (numOfClasses - 1.0) * log(numOfVertices)
			- 0.5 * (numOfClasses + 3.0)
					* log((numOfVertices * (numOfVertices - 1.0)) / 2.0);
	return ICL;
}

void MMExcessTrustModel::updateKappa() {
	// update gamma first
	updateGamma();
	// now update gamma by summing over components
	updateIndependentKappa();
}

void MMExcessTrustModel::updateIndependentKappa() {
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++)
			kappa[k][l] = log(gamma[k][l]);
}

void MMExcessTrustModel::updateGamma() {
	// update gamma components first
	updateGamma1();
	updateGamma2();
	updateGamma3();
	updateGamma4Minus();
	updateGamma4Plus();
	// now update gamma by summing over components
	updateIndependentGamma();
}

void MMExcessTrustModel::updateIndependentGamma() {
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++)
			gamma[k][l] = 1 + gamma1[k][l] + gamma1[l][k] + gamma2[k][l]
					+ gamma2[l][k] + gamma3[k][l] + gamma3[l][k]
					+ gamma4Minus[k][l] + gamma4Plus[k][l];
}

void MMExcessTrustModel::updateGamma1() {
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++)
			gamma1[k][l] = exp(negDenTheta - excessTrustTheta[l]);
}

void MMExcessTrustModel::updateGamma2() {
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++)
			gamma2[k][l] = exp(excessTrustTheta[l]);
}

void MMExcessTrustModel::updateGamma3() {
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++)
			gamma3[k][l] = exp(
					negDenTheta + excessTrustTheta[k] - excessTrustTheta[l]);
}

void MMExcessTrustModel::updateGamma4Minus() {
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++)
			gamma4Minus[k][l] = exp(
					2 * negDenTheta + negRecTheta - excessTrustTheta[k]
							- excessTrustTheta[l]);
}

void MMExcessTrustModel::updateGamma4Plus() {
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++)
			gamma4Plus[k][l] = exp(
					posRecTheta + excessTrustTheta[k] + excessTrustTheta[l]);
}

void MMExcessTrustModel::copyParameters(DegreeModel* model) {
	MMExcessTrustModel* _model = dynamic_cast<MMExcessTrustModel*>(model);
	// Class mixing probabilities;
	alpha = _model->alpha;
	// Theta paramters
	excessTrustTheta = _model->excessTrustTheta;
	negDenTheta = _model->negDenTheta;
	negRecTheta = _model->negRecTheta;
	posRecTheta = _model->posRecTheta;
	// Membership variational parameters
	tau = _model->tau;
}

void MMExcessTrustModel::writeModel2File(ofstream & modelFile) {
	// Class mixing probabilities;
	modelFile << "alpha:" << endl;
	printDoubleVector2TextFile(alpha, modelFile);
	modelFile << endl;
	// Degree paramters
	modelFile << "excessTrustTheta:" << endl;
	printDoubleVector2TextFile(excessTrustTheta, modelFile);
	modelFile << endl;
	modelFile << "negDenTheta:" << endl;
	modelFile << negDenTheta << endl;
	modelFile << "posDenTheta:" << endl;
	modelFile << "negRecTheta:" << endl;
	modelFile << negRecTheta << endl;
	modelFile << "posRecTheta:" << endl;
	modelFile << posRecTheta << endl;
	// Membership variational parameters
	modelFile << "tau:" << endl;
	for (unsigned int i = 0; i < tau.size() - 1; i++) {
		printDoubleVector2TextFile(tau[i], modelFile);
		modelFile << endl;
	}
	printDoubleVector2TextFile(tau[tau.size() - 1], modelFile);
}

void MMExcessTrustModel::writeParameters2File(ofstream & modelFile) {
	// Class mixing probabilities;
	printDoubleVector2TextFile(alpha, modelFile);
	modelFile << endl;
	// Degree paramters
	printDoubleVector2TextFile(excessTrustTheta, modelFile);
	modelFile << endl;
	modelFile << negDenTheta << endl;
	modelFile << negRecTheta << endl;
	modelFile << posRecTheta << endl;
}

}
