/*
 * NRMMExcessTrustModel.cpp
 *
 *  Created on: Mar 9, 2012
 *      Author: duyvu
 */

#include "NRMMExcessTrustModel.h"

namespace mixderg {

NRMMExcessTrustModel::NRMMExcessTrustModel() {

}

NRMMExcessTrustModel::~NRMMExcessTrustModel() {
	// clean memory
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++)
			if (dyadProbs[k][l] != NULL)
				delete dyadProbs[k][l];
}

void NRMMExcessTrustModel::setNumOfClasses(int _numOfClasses) {
	numOfClasses = _numOfClasses;

	// constants
	numDyads = 9;
	numStats = 2 + 3;
	numParams = numOfClasses + 3;

	// create the table of dyad value probabilities
	dyadProbs = DoublePointerMatrix(numOfClasses,
			DoublePointerVector(numOfClasses));
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++)
			dyadProbs[k][l] = new double[numDyads];

	// create the transform matrix
	A_t = Matrix(numParams, numStats);
	A_t.set(0); // reset to zero
	for (unsigned int k = 0; k < (numParams - numOfClasses); k++)
		A_t[numOfClasses + k][2 + k] = 1; // global parameters

	// initialize the test table of statistics and dyad values
	statDyadTest = Matrix(numStats, numDyads);
	statDyadTest.set(0);
	// index:   0, 1, 2, 3, 4, 5, 6, 7, 8
	// y_ij = [-1, 0, 1, 0,-1, 1,-1, 1, 0];
	// y_ji = [ 0,-1, 0, 1, 1,-1,-1, 1, 0];

	// y_ji
	statDyadTest[0][1] = -1;
	statDyadTest[0][3] = 1;
	statDyadTest[0][4] = 1;
	statDyadTest[0][5] = -1;
	statDyadTest[0][6] = -1;
	statDyadTest[0][7] = 1;

	// y_ij
	statDyadTest[1][0] = -1;
	statDyadTest[1][2] = 1;
	statDyadTest[1][4] = -1;
	statDyadTest[1][5] = 1;
	statDyadTest[1][6] = -1;
	statDyadTest[1][7] = 1;

	// y_ij == -1 + y_ji == -1
	statDyadTest[2][0] = 1;
	statDyadTest[2][1] = 1;
	statDyadTest[2][4] = 1;
	statDyadTest[2][5] = 1;
	statDyadTest[2][6] = 2;

	// y_ij == y_ji == -1
	statDyadTest[3][6] = 1;

	// y_ij == y_ji == +1
	statDyadTest[4][7] = 1;
}

void NRMMExcessTrustModel::runModelEstimationMStep(int emIter) {

	cout << "NRMMExcessTrustModel::runModelEstimationMStep at " << emIter
			<< endl;

	// alpha
	prevAlpha.swap(alpha);
	sumDoubleMatrixOverRows(tau, alpha);
	for (unsigned int k = 0; k < alpha.size(); k++)
		alpha[k] /= numOfVertices;
	normalizeVector(alpha, minAlpha);

	// theta
	DoubleVector params(numParams);

	// copy current parameters to the param vector
	copyThetas2Params(params);

	// optimize the param vector
	runNewtonRaphson(params, 100, 1e-20, 1e-10);

	// store previous parameters
	prev_excessTrustTheta.swap(excessTrustTheta);
	prev_negDenTheta = negDenTheta;
	prev_negRecTheta = negRecTheta;
	prev_posRecTheta = posRecTheta;

	// get new parameter estimates
	copyParams2Thetas(params);
	// update caching parameters
	updateKappa();
}

void NRMMExcessTrustModel::copyThetas2Params(DoubleVector & params) {
	// excess trust parameters
	for (unsigned int k = 0; k < numOfClasses; k++)
		params[k] = excessTrustTheta[k]; // excess trust
	// negative density parameter
	params[numOfClasses] = negDenTheta;
	// negative reciprocity parameter
	params[numOfClasses + 1] = negRecTheta;
	// positive reciprocity parameter
	params[numOfClasses + 2] = posRecTheta;
}

void NRMMExcessTrustModel::copyParams2Thetas(const DoubleVector & params) {
	// excess trust parameters
	for (unsigned int k = 0; k < numOfClasses; k++)
		excessTrustTheta[k] = params[k]; // excess trust
	// negative density parameter
	negDenTheta = params[numOfClasses];
	// negative reciprocity parameter
	negRecTheta = params[numOfClasses + 1];
	// positive reciprocity parameter
	posRecTheta = params[numOfClasses + 2];
}

// 100, 1e-20, 1e-10
void NRMMExcessTrustModel::runNewtonRaphson(DoubleVector& params, int maxSteps,
		double relLLHTolerance, double scoreTolerance) {

	Vec u(numParams);
	Matrix I(numParams, numParams);
	Vec dir(numParams);
	int* ipvt = ivector(numParams);

	double currentLogLLH = getLogLLH(params);

	for (int k = 0; k < maxSteps; k++) {

		// obtain the score vector and the Hessian matrix
		getScoreNegHessian(params, u, I);

		// compute the moving direction dir = I^-1 * u
		ludcmp(I, ipvt);
		solvlu(I, u, dir, ipvt);

		// search for a good step length
		double stepLength = 1.0;
		DoubleVector newParams(numParams);
		// compute new estimates
		for (int h = 0; h < numParams; h++)
			newParams[h] = params[h] + stepLength * dir[h];
		// compute new Log LLH
		double newLogLLH = getLogLLH(newParams);

		// number of trials for the step length
		unsigned searchCount = 1;
		while (isnan(newLogLLH) || newLogLLH < currentLogLLH) {
			// let halve the step length
			stepLength /= 2.0;
			// compute new estimates
			for (int h = 0; h < numParams; h++)
				newParams[h] = params[h] + stepLength * dir[h];
			// compute new Log LLH
			newLogLLH = getLogLLH(newParams);
			// one more trial
			searchCount++;
		}

		// update params
		params = newParams;

		// compute the largest component of the score vector
		double maxError = fabs(u[0]);
		for (int l = 1; l < numParams; l++)
			if (maxError <= fabs(u[l]))
				maxError = fabs(u[l]);

		// compute the relative improvement of Log LLH
		double relLogLLHImprovement = fabs(
				(currentLogLLH - newLogLLH) / currentLogLLH);

		// update currentlogLLH
		currentLogLLH = newLogLLH;

		// check the score vector: stop if all of elements are smaller than the tolerance
		if (fabs(maxError) < scoreTolerance)
			break;

		// check if the relative improvement of Log LLH is small enough to stop
		if (relLogLLHImprovement < relLLHTolerance)
			break;
	}

	// cleaning
	free(ipvt);
}

void NRMMExcessTrustModel::copyParams2Thetas(const DoubleVector & params,
		DoubleVector & excessTrustTheta, double & negDenTheta,
		double & negRecTheta, double & posRecTheta) {
	// excess trust parameters
	for (unsigned int k = 0; k < numOfClasses; k++)
		excessTrustTheta[k] = params[k]; // excess trust
	// negative density parameter
	negDenTheta = params[numOfClasses];
	// negative reciprocity parameter
	negRecTheta = params[numOfClasses + 1];
	// positive reciprocity parameter
	posRecTheta = params[numOfClasses + 2];
}

void NRMMExcessTrustModel::computeCachingParameters(
		const DoubleVector & excessTrustTheta, const double & negDenTheta,
		const double & negRecTheta, const double & posRecTheta,
		DoubleMatrix & gammaMinus0, DoubleMatrix & gammaPlus0,
		DoubleMatrix & gammaMinusPlus, DoubleMatrix & gammaRecMinus,
		DoubleMatrix & gammaRecPlus, DoubleMatrix & gamma,
		DoubleMatrix & kappa) {
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++) {
			gammaMinus0[k][l] = exp(negDenTheta - excessTrustTheta[l]);
			gammaPlus0[k][l] = exp(excessTrustTheta[l]);
			gammaMinusPlus[k][l] = exp(
					negDenTheta + excessTrustTheta[k] - excessTrustTheta[l]);
			gammaRecMinus[k][l] = exp(
					2 * negDenTheta + negRecTheta - excessTrustTheta[k]
							- excessTrustTheta[l]);
			gammaRecPlus[k][l] = exp(
					posRecTheta + excessTrustTheta[k] + excessTrustTheta[l]);
		}
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++) {
			gamma[k][l] = 1 + gammaMinus0[k][l] + gammaMinus0[l][k]
					+ gammaPlus0[k][l] + gammaPlus0[l][k] + gammaMinusPlus[k][l]
					+ gammaMinusPlus[l][k] + gammaRecMinus[k][l]
					+ gammaRecPlus[k][l];
			kappa[k][l] = log(gamma[k][l]);
		}
}

double NRMMExcessTrustModel::getLogLLH(const DoubleVector& params) {

	// original parameters
	DoubleVector excessTrustTheta = DoubleVector(numOfClasses);
	double negDenTheta, negRecTheta, posRecTheta;

	copyParams2Thetas(params, excessTrustTheta, negDenTheta, negRecTheta,
			posRecTheta);

	// original caching parameters
	DoubleMatrix gammaMinus0 = DoubleMatrix(numOfClasses,
			DoubleVector(numOfClasses));
	DoubleMatrix gammaPlus0 = DoubleMatrix(numOfClasses,
			DoubleVector(numOfClasses));
	DoubleMatrix gammaMinusPlus = DoubleMatrix(numOfClasses,
			DoubleVector(numOfClasses));
	DoubleMatrix gammaRecMinus = DoubleMatrix(numOfClasses,
			DoubleVector(numOfClasses));
	DoubleMatrix gammaRecPlus = DoubleMatrix(numOfClasses,
			DoubleVector(numOfClasses));
	DoubleMatrix gamma = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));
	DoubleMatrix kappa = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));

	computeCachingParameters(excessTrustTheta, negDenTheta, negRecTheta,
			posRecTheta, gammaMinus0, gammaPlus0, gammaMinusPlus, gammaRecMinus,
			gammaRecPlus, gamma, kappa);

	double llh = 0;

	// the linear combination of statistics and coefficients

	// tau * excess trust statistics
	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int k = 0; k < numOfClasses; k++)
			llh += tau[i][k] * (excessTrustTheta[k] * excessTrustDegrees[i]);
	// density statistics
	llh += negDenTheta * negDensity;
	// reciprocity statistics
	llh += negRecTheta * negReciprocity;
	llh += posRecTheta * posReciprocity;

	// the normalizing constant
	llh -= computeCarryOverTerm(kappa);

	return llh;
}

double NRMMExcessTrustModel::computeAlphaProduct(unsigned int k,
		unsigned int l) {
	double alphaTerm = 0;

	double carryTerm = 0;
	for (unsigned int i = 1; i < numOfVertices; i++) {
		carryTerm += tau[i - 1][l];
		alphaTerm += tau[i][k] * carryTerm;
	}

	return alphaTerm;
}

void NRMMExcessTrustModel::computeDyadProbs(const DoubleMatrix & gammaMinus0,
		const DoubleMatrix & gammaPlus0, const DoubleMatrix & gammaMinusPlus,
		const DoubleMatrix & gammaRecMinus, const DoubleMatrix & gammaRecPlus,
		const DoubleMatrix & gamma) {
	// probabilities of dyad values by classes
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++) {
			// y_ij = [-1,0,1,0,-1,1,-1,1,0];
			// y_ji = [0,-1,0,1,1,-1,-1,1,0];
			// p0(k,l),p0(l,k),p1(k,l),p1(l,k),p2(k,l),p2(l,k),p3(k,l),p4(k,l),p5(k,l)
			dyadProbs[k][l][0] = gammaMinus0[k][l] / gamma[k][l];
			dyadProbs[k][l][1] = gammaMinus0[l][k] / gamma[k][l];
			dyadProbs[k][l][2] = gammaPlus0[k][l] / gamma[k][l];
			dyadProbs[k][l][3] = gammaPlus0[l][k] / gamma[k][l];
			dyadProbs[k][l][4] = gammaMinusPlus[k][l] / gamma[k][l];
			dyadProbs[k][l][5] = gammaMinusPlus[l][k] / gamma[k][l];
			dyadProbs[k][l][6] = gammaRecMinus[k][l] / gamma[k][l];
			dyadProbs[k][l][7] = gammaRecPlus[k][l] / gamma[k][l];
			dyadProbs[k][l][8] = 1.0 / gamma[k][l];
		}
}

void NRMMExcessTrustModel::computeStatMean(Vec & statMean, unsigned int k,
		unsigned int l) {
	for (int statIndex = 0; statIndex < numStats; statIndex++) {
		statMean[statIndex] = 0;
		for (int dyadIndex = 0; dyadIndex < numDyads; dyadIndex++)
			statMean[statIndex] += statDyadTest[statIndex][dyadIndex]
					* dyadProbs[k][l][dyadIndex];
	}
}

void NRMMExcessTrustModel::computeStatCov(Matrix & statCov, unsigned int k,
		unsigned int l) {
	for (int statIndex1 = 0; statIndex1 < numStats; statIndex1++)
		for (int statIndex2 = 0; statIndex2 < numStats; statIndex2++) {
			statCov[statIndex1][statIndex2] = 0;
			for (int dyadIndex = 0; dyadIndex < numDyads; dyadIndex++)
				statCov[statIndex1][statIndex2] +=
						statDyadTest[statIndex1][dyadIndex]
								* statDyadTest[statIndex2][dyadIndex]
								* dyadProbs[k][l][dyadIndex];
		}
}

void NRMMExcessTrustModel::getScoreNegHessian(const DoubleVector & params,
		Vec & u, Matrix & I) {

	// original parameters
	DoubleVector excessTrustTheta = DoubleVector(numOfClasses);
	double negDenTheta, negRecTheta, posRecTheta;

	copyParams2Thetas(params, excessTrustTheta, negDenTheta, negRecTheta,
			posRecTheta);

	// original caching parameters
	DoubleMatrix gammaMinus0 = DoubleMatrix(numOfClasses,
			DoubleVector(numOfClasses));
	DoubleMatrix gammaPlus0 = DoubleMatrix(numOfClasses,
			DoubleVector(numOfClasses));
	DoubleMatrix gammaMinusPlus = DoubleMatrix(numOfClasses,
			DoubleVector(numOfClasses));
	DoubleMatrix gammaRecMinus = DoubleMatrix(numOfClasses,
			DoubleVector(numOfClasses));
	DoubleMatrix gammaRecPlus = DoubleMatrix(numOfClasses,
			DoubleVector(numOfClasses));
	DoubleMatrix gamma = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));
	DoubleMatrix kappa = DoubleMatrix(numOfClasses, DoubleVector(numOfClasses));

	computeCachingParameters(excessTrustTheta, negDenTheta, negRecTheta,
			posRecTheta, gammaMinus0, gammaPlus0, gammaMinusPlus, gammaRecMinus,
			gammaRecPlus, gamma, kappa);

	computeDyadProbs(gammaMinus0, gammaPlus0, gammaMinusPlus, gammaRecMinus,
			gammaRecPlus, gamma);

	// clear gradient and Hessian
	u.set(0);
	I.set(0);

	// gradient left
	for (unsigned int classIndex = 0; classIndex < numOfClasses; classIndex++)
		u[classIndex] = computeWeightedDegrees(classIndex, excessTrustDegrees);
	u[numOfClasses] = negDensity;
	u[numOfClasses + 1] = negReciprocity;
	u[numOfClasses + 2] = posReciprocity;

	// gradient right
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++) {
			// set A_t
			A_t[k][0] = 1;
			A_t[l][1] = 1;

			// compute E[g_kl(D)]
			Vec statMean(numStats);
			computeStatMean(statMean, k, l);

			// compute E[g_kl(D) * g_kl(D)^T]
			Matrix statCov(numStats, numStats);
			computeStatCov(statCov, k, l);

			// compute sum of alpha_ik * alpha_jl
			double alphaTerm_k_l = computeAlphaProduct(k, l);

			// update the gradient vector
			u = u - A_t * (statMean * alphaTerm_k_l);

			// update the Hessian matrix
			Matrix addedMatrix = A_t * (statCov * A_t.transpose());
			I.addScaledMatrix(addedMatrix, alphaTerm_k_l);

			// reset A_t
			A_t[k][0] = 0;
			A_t[l][1] = 0;
		}
}

}
