/*
 * signedNetworks.h
 *
 *  Created on: Oct 12, 2011
 *      Author: duyvu
 */

#ifndef SIGNEDNETWORKS_H_
#define SIGNEDNETWORKS_H_

#include "em/EMEngine.h"

#include "var/ParametricBoostrapSampler.h"
#include "var/signed/EfficientExcessTrustParametricBoostrapSampler.h"

#define MATHLIB_STANDALONE
#include <Rmath.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <ctime>
#include <string>
#include<sys/stat.h>
#include<sys/types.h>

using namespace std;
using namespace mixderg;

int estimateParameters(const char* confFile, const char* outputDir,
		unsigned int numOfClusters, unsigned int fromRun, unsigned int toRun) {

	cout << "Estimating a Signed Degree Model WITHOUT convergence details"
			<< endl;
	cout << "confFile is " << confFile << endl;
	cout << "outputDirectory is " << outputDir << endl;
	cout << "Number of classes is " << numOfClusters << endl;

	EMEngine engine = EMEngine();
	engine.configure(confFile, "\t");
	engine.setNumOfClasses(numOfClusters);
	engine.setNumOfEMRuns(toRun - fromRun + 1);
	set_seed(fromRun, toRun);

	string outputDirectory = string(outputDir);
	unsigned int outputPrecision = 10;

	stringstream postfix;
	postfix << engine.getNumOfClasses() << "_";
	postfix << engine.getMaxNumOfEMIterations() << "_";
	postfix << engine.getMaxNumOfESteps();

	string modelICLFilename = outputDirectory + "/modelICL_" + postfix.str()
			+ ".txt";

	string bestModelFilename = outputDirectory + "/bestModel_" + postfix.str()
			+ ".txt";

	string bestICLFilename = outputDirectory + "/bestICL_" + postfix.str()
			+ ".txt";

	engine.runOneModel(modelICLFilename, bestModelFilename, bestICLFilename,
			outputPrecision, fromRun);

	return 0;
}

int estimateParametersWithCovergenceDetails(const char* confFile,
		const char* outputDir, unsigned int numOfClusters, unsigned int fromRun,
		unsigned int toRun) {

	cout << "Estimating a Signed Degree Model WITH convergence details" << endl;
	cout << "confFile is " << confFile << endl;
	cout << "outputDirectory is " << outputDir << endl;
	cout << "Number of classes is " << numOfClusters << endl;

	EMEngine engine = EMEngine();
	engine.configure(confFile, "\t");
	engine.setNumOfClasses(numOfClusters);
	engine.setNumOfEMRuns(toRun - fromRun + 1);
	set_seed(fromRun, toRun);

	string outputDirectory = string(outputDir);
	unsigned int outputPrecision = 10;

	stringstream postfix;
	postfix << engine.getNumOfClasses() << "_";
	postfix << engine.getMaxNumOfEMIterations() << "_";
	postfix << engine.getMaxNumOfESteps();

	string modelICLFilename = outputDirectory + "/modelICL_" + postfix.str()
			+ ".txt";

	string bestModelFilename = outputDirectory + "/bestModel_" + postfix.str()
			+ ".txt";

	string bestICLFilename = outputDirectory + "/bestICL_" + postfix.str()
			+ ".txt";

	engine.runDetailedOneModel(modelICLFilename, bestModelFilename,
			bestICLFilename, outputPrecision, fromRun);

	return 0;
}

int sampleSignedNetworksForStdErrors(unsigned int fromSampleIndex,
		unsigned int toSampleIndex, const char* modelFilename,
		const char* confFile, const char* outputDir) {

	set_seed(fromSampleIndex, toSampleIndex);

	ParametricBoostrapSampler* sampler =
			new EfficientExcessTrustParametricBoostrapSampler();

	cout << "Setting parameters!" << endl;
	sampler->setParameters(modelFilename, "\t");

	cout << "Setting configuration!" << endl;
	sampler->configure(confFile, "\t");

	cout << "Bootstrapping samples!" << endl;
	string sampleDir = string(outputDir) + "/samples";
	if (mkdir(sampleDir.c_str(), 0777) == -1)
		exit(1);
	sampler->generateSamples(fromSampleIndex, toSampleIndex, sampleDir.c_str(),
			10);

	cout << "Estimating samples!" << endl;
	string estimateDir = string(outputDir) + "/estimates";
	if (mkdir(estimateDir.c_str(), 0777) == -1)
		exit(1);

	// Using pre-specified iterations in the configuration file
	sampler->estimateSamples(fromSampleIndex, toSampleIndex, sampleDir,
			estimateDir, 10, true);

	// Using stopping criteria in the configuration file
	// sampler->estimateSamples(fromSampleIndex, toSampleIndex, sampleDir,
	//		estimateDir, 10, false);

	if (sampler != NULL)
		delete sampler;

	return 0;
}

int runSignedNetworkExperiments(int argc, char *argv[]) {

	switch (atoi(argv[1])) {
	case 0:
		// const char* confFile: the path to the configuration file
		// const char* outputDir: the path to the directory where the output should be
		// unsigned int numOfClusters: the number of clusters
		// unsigned int fromRun: the beginning run number which is used for setting random seed 1
		// unsigned int toRun: the ending run number which is used for setting random seed 2
		// The last two parameters allow for paralleling the runs on cluster computers.
		// For example, we can specify 1 to 10, and 11 to 20 to cut 20 runs into 2 parallel sections.
		if (argc > 6)
			return estimateParameters(argv[2], argv[3], atoi(argv[4]),
					atoi(argv[5]), atoi(argv[6]));
		break;
	case 1:
		// const char* confFile: the path to the configuration file
		// const char* outputDir: the path to the directory where the output should be
		// unsigned int numOfClusters: the number of clusters
		// unsigned int fromRun: the beginning run number which is used for setting random seed 1
		// unsigned int toRun: the ending run number which is used for setting random seed 2
		// The last two parameters allow for paralleling the runs on cluster computers.
		// For example, we can specify 1 to 10, and 11 to 20 to cut 20 runs into 2 parallel sections.
		if (argc > 6)
			return estimateParametersWithCovergenceDetails(argv[2], argv[3],
					atoi(argv[4]), atoi(argv[5]), atoi(argv[6]));
		break;
	case 2:
		// unsigned int fromSampleIndex: the beginning sample index which is used to set the random seed 1
		// unsigned int toSampleIndex: the ending sample index which is used to set the random seed 2
		// const char* modelFilename: the path to the model parameter file
		// const char* confFile: the path to the configuration file
		// const char* outputDir: the path to the directory where the output should be
		// The first two parameters allow for paralleling the runs on cluster computers.
		if (argc > 6)
			return sampleSignedNetworksForStdErrors(atoi(argv[2]),
					atoi(argv[3]), argv[4], argv[5], argv[6]);
		break;
	default:
		cout << "Please enter another first parameter!!!" << endl;
		cout
				<< "0: Run the EM method to estimate parameters WITHOUT convergence details"
				<< endl;
		cout
				<< "1: Run the EM method to estimate parameters WITH convergence details"
				<< endl;
		cout << "2: Run the bootstrap method to estimate standard errors"
				<< endl;
		break;
	}

	return 0;
}

#endif /* SIGNEDNETWORKS_H_ */
